########################################################################
# OSM::Tree::Relation
# OSM::Tree::Member

=encoding utf8

=head1 NAME

OSM::Tree::Relation -- OSM Relation

=head1 SYNOPSIS


=head1 DESCRIPTION

the Relation-Object
derived from L<OSM::Tree::BASE>

=head1 AUTHOR

ToPeG

=head2 OSM::Tree::Relation

the relation-object

=over 4

=cut

package OSM::Tree::Relation;
use strict;
use warnings;
use base 'OSM::Tree::BASE';

sub _new
{
  my $class=shift;
  my $self={};
  $self->{TREE}=shift;
  $self->{ID}=shift;
  $self->{IGNORE_DELETED}=shift // \do{my $t=0};
  $self->{FORCE_DELETE}=shift  // \do{my $t=0};
  return undef unless(exists($self->{TREE}->{relation}->{$self->{ID}}));
  $self->{REF}=$self->{TREE}->{relation}->{$self->{ID}};
  bless($self,$class);
  return $self;
}

#=======================================================================

=item member(ID)

get an member of this relation

returns an member-object
or undef if there is no member with this id

=cut
sub member
{
  my $self=shift;
  my $id=shift;
  if(exists($self->{REF}->{member}->{$id}))
  {
    my $mref=$self->{REF}->{member}->{$id};
    my $child=undef;

    if($mref->{type} eq 'way')
    { $child=OSM::Tree::Way->_new($self->{TREE},$id,$self->{IGNORE_DELETED},$self->{FORCE_DELETE}); }
    elsif($mref->{type} eq 'node')
    { $child=OSM::Tree::Node->_new($self->{TREE},$id,$self->{IGNORE_DELETED},$self->{FORCE_DELETE}); }
    elsif($mref->{type} eq 'relation')
    {  $child=OSM::Tree::Relation->_new($self->{TREE},$id,$self->{IGNORE_DELETED},$self->{FORCE_DELETE}); }

    return OSM::Tree::Member->_new($mref->{type}, $child, $mref->{role}) if(defined($child));
  }
  return wantarray?():undef;
}

=item member_add(<HASH>)

add member

keys are:

- ref    =   ID of the Way/Node/Relation
- type   =   type of the member (way/node/relation)
- role   =   The Role

see L<OSM::Tree::Member|OSM::Tree::Relation/OSM::Tree::Member> and OSM-XML spec for more

returns the an member-object if the ref points to an Object in the OSM::Tree
or 1 if not or return undef if it fails;

=cut
sub member_add
{
  my $self=shift;
  my $opts=$_[0] if($_[0] && ref($_[0]) eq 'HASH');
  unless($opts)
  {
    pop(@_) if(@_%2);
    $opts={@_};
  }

  my $id=delete($opts->{ref});
  return undef if(!defined($id) || $id=~/\D/);
  return undef if(exists($self->{REF}->{member}->{$id}));
  return undef if(!$opts->{type} || !grep{$_ eq $opts->{type}}qw(way node relation));
  $self->{REF}->{member}->{$id} = { ref=>$id, type=>$opts->{type}, role=>$opts->{role} };
  my $ret=$self->member($id);
  $ret=1 unless($ret);
  $self-> _set_modify();
  return $ret;
}

=item member_del(MEMBER)

remove a member
returns 1 on success;

=cut
sub member_del
{
  my $self=shift;
  my $member=shift;
  return 0 unless(defined($member) && $member->isa('OSM::Tree::Member'));
  return $self->member_id_del($member->child->id());
}

=item member_id_del(ID)

remove a member
returns 1 on success;

=cut
sub member_id_del
{
  my $self=shift;
  my $id=shift;
  return 0 if(!defined($id) || $id=~/\D/);
  return 0 if(!exists($self->{REF}->{member}->{$id}));
  delete($self->{REF}->{member}->{$id});
  $self-> _set_modify();
  return 1;
}


=item member_list(<CODE>)

get an member-object-list if this relation
if an coderefence is given it will use as filter.

returns an member-object-list
or an empty list if there are no members

=cut
sub member_list
{
  my $self=shift;
  my $filter=shift;
  my @l=();
  if($filter)
  {
    if(ref($filter) eq 'CODE')
    {
      for($self->member_id_list())
      {
        my $m=$self->member($_);
        push(@l,$m) if($m && $filter->($m));
      }
    }
    else
    { croak(q(Given filter is not a CodeRefence!)); }
  }
  else
  {
    for($self->member_id_list())
    {
      my $m=$self->member($_);
      push(@l,$m) if($m);
    }
  }
  return @l;
}

=item member_id_list()

get an member-id-list if this relation

returns an member-id-list
or an empty list if there are no members

=cut
sub member_id_list
{
  my $self=shift;
  return keys(%{$self->{REF}->{member}});
}

=item member_list_length()

get the length the list of all members

returns the length of the member-list

=cut
sub member_list_length
{ return scalar(shift->member_id_list()); }

=item member_exists(ID)

returns true if the member-id exists

=cut
sub member_exists
{
  my $self=shift;
  my $nid=shift;
  return 0 unless(defined($nid));
  if(ref($nid) && ref($nid) eq 'OSM::Tree::Member')
  { $nid=$nid->child()->id(); }
  return 1 if(exists($self->{REF}->{member}->{$nid}));
  return 0;
}

=item kind()

returns $self->tag('type') but is faster.

=cut
sub kind
{
  my $self=shift;
  $self->{REF}->{tag}->{type}='' unless($self->{REF}->{tag}->{type});
  return $self->{REF}->{tag}->{type};
}

#-----------------------------------------------------------------------
# overwrite

sub to_xml
{
  my $self=shift;
  my $in=shift || '';
  my $ref=$self->{REF};
  $ref->{visible} ||= 'true';
  my @vals;
  for my $name (qw(id version timestamp uid user visible changeset))
  { push(@vals,qq($name="$ref->{$name}")) if(defined($ref->{$name})); }
  my $txt=join(' ',@vals);
  if(exists($ref->{action}) && $ref->{action})
  { $txt.=qq( action="$ref->{action}"); }
  my $ret=sprintf(qq!$in<relation %s >\n!,$txt);
  $ret.=$_->to_xml("$in  ") for(sort $self->member_list());
  #$ret  .=$_->to_xml("$in  ") for(sort{$a->child->id() <=> $b->child->id()}$self->member_list());
  $ret  .=$self->tags()->to_xml("$in  ") if($self->has_tags());
  $ret  .="$in</relation>\n";
  return $ret;
}

########################################################################

=back

=head2 OSM::Tree::Member

the member-object

=over 4

=cut

package OSM::Tree::Member;

sub _new
{
  my $class=shift;
  my $self={};
  $self->{TYPE}=shift;
  $self->{CHILD}=shift;
  $self->{ROLE}=shift;
  bless($self,$class);
  return $self;
}

=item type()

get the type of the child of this member

returns an String of the type ("way","node","relation")

=cut
sub type
{ return shift->{TYPE}; }

=item child()

get the child of this member

returns an Object

=cut
sub child
{ return shift->{CHILD}; }

=item role()

returns a String or undef if the XML-tag ist not set

=cut
sub role
{ return shift->{ROLE}; }

sub to_xml
{
  my $self=shift;
  my $in=shift || '';
  my $type=$self->{TYPE} || '';
  my $ref=$self->{CHILD}? $self->{CHILD}->id(): '';
  my $role=$self->{ROLE} || '';
  return qq!$in<member type="$type" ref="$ref" role="$role"/>\n!;
}
1;


=back

=head1 See also

L<OSM::Tree::BASE>

L<OSM::Tree::Node>

L<OSM::Tree::Way>

L<OSM::Tree::Relation>

L<OSM::Tree::TagList>

L<OSM::Tree::Polygon>

L<OSM::Tree>

=cut
