########################################################################
# OSM::Tree::TagList
# OSM::Tree::Tag

=encoding utf8

=head1 NAME

OSM::Tree::TagList -- OSM Node

=head1 SYNOPSIS


=head1 DESCRIPTION

Parse OpenStreetMap XML file

=head1 AUTHOR

ToPeG

=head2 OSM::Tree::TagList

The TagList Object

=over 4

=cut

package OSM::Tree::TagList;
use strict;
use warnings;

sub _new
{
  my $class=shift;
  my $self={};
  $self->{ELEMENT}=shift;
  $self->{TAGS}=shift;
  $self->{INSENSITIVE}=shift;
  $self->{TAGS}={} unless($self->{TAGS} && ref($self->{TAGS}) eq 'HASH');
  bless($self,$class);
  return $self;
}

#=======================================================================

=item element()

returns the parent object, of this TagList-object;
can be: node, way or relation

=cut
sub element
{ return shift->{ELEMENT}; }

=item case_sensetive(<1/0>)

get/set the casesensitivity of catching keys
returns true/false

=cut
sub case_sensetive
{
  my $self=shift;
  my $case=shift;
  $self->{INSENSITIVE}=$case if(defined($case));
  return $self->{INSENSITIVE};
}

=item case(<1/0>)

alias for case_sensetive

=cut
sub case
{ return shift->case_sensetive(@_); }

=item key_list([@list])

returns a List of the Keys in this TagList
if an list of names ist set,
only Names in the list will be returned

=cut
sub key_list
{
  my $self=shift;
  my @list=@_;
  if(@list)
  {
    my @ret;
    for my $key (@list)
    { push(@ret,$key) if($self->has_key($key)); }
    return @ret;
  }
  return keys(%{$self->{TAGS}});
}

=item key_list_length()

returns the length of this TagList

=cut
sub key_list_length
{ return scalar(shift->key_list()); }

=item key_match(RegExp)

returns a list of Tags where the Keys matchs the ReguarExpression

=cut
sub key_match
{
  my $self=shift;
  my $regexp=shift;
  my @ret;
  if($self->{INSENSITIVE})
  {
    for my $key ($self->key_list())
    {
      push(@ret,$self->tag($key)) if($key=~/$regexp/i);
    }
  }
  else
  {
    for my $key ($self->key_list())
    {
      push(@ret,$self->tag($key)) if($key=~/$regexp/);
    }
  }
  return @ret;
}

=item has_key($name)

return true if the given TagName exists

=cut
sub has_key
{
  my $self=shift;
  my $name=shift;
  return $self->key_list_length()>0?1:0 unless defined($name);
  if($self->{INSENSITIVE})
  {
    for my $key ($self->key_list())
    { return 1 if(lc($key) eq lc($name)); }
    return 0;
  }
  return 1 if exists($self->{TAGS}->{$name});
  return 0;
}

=item has_all_keys($name1,$name2)

return true if all the given TagNames exists

=cut
sub has_all_keys
{
  my $self=shift;
  my @names=@_;
  return 1 if(scalar(@names) == scalar($self->key_list(@names)));
  return 0;
}
#=======================================================================

=item tag(KEY)

return the tag-Object from this key

=cut
sub tag
{
  my $self=shift;
  my $key=shift;
  if($self->{INSENSITIVE})
  {
    for my $skey ($self->key_list())
    {
      if(lc($skey) eq lc($key))
      {
        return OSM::Tree::Tag->_new($self->{ELEMENT},$skey,$self->{TAGS}->{$skey});
      }
    }
    return OSM::Tree::TagDummy->_new($self->{ELEMENT},$key);
  }
  return OSM::Tree::TagDummy->_new($self->{ELEMENT},$key) unless($key && exists($self->{TAGS}->{$key}));
  return OSM::Tree::Tag->_new($self->{ELEMENT},$key,$self->{TAGS}->{$key});
}


=item tag_add([KEY,VALUE] | TAG)

Create new Tag-Objekt
return the tag-Object or undef

=cut
sub tag_add
{
  my $self=shift;
  my $key=shift;
  my $value=shift;
  if(defined($key) && ref($key) && $key->isa('OSM::Tree:Tag'))
  {
    $value=$key->value();
    $key=$key->name();
  }
  return undef unless($key && defined($value));
  return undef if($self->has_tag($key));
  $self->{TAGS}->{$key}=$value;
  $self->element->_set_modify();
  return $self->tag($key);
}

=item tag_del(KEY|TAG)

delete a tag
returns 1 on success

=cut
sub tag_del
{
  my $self=shift;
  my $key=shift;
  return 0 if(!$key);
  $key=$key->name() if(ref($key) && $key->isa('OSM::Tree::Tag'));
  return 0 if(!exists($self->{TAGS}->{$key}));
  delete($self->{TAGS}->{$key});
  $self->element->_set_modify();
  return 1;
}

=item tag_list([@names])

returns a List of the Tags in this TagList
if an list of names ist set,
only tags with a Name in the list will be returned

=cut
sub tag_list
{
  my $self=shift;
  my @ret=$self->key_list(@_);
  $_=$self->tag($_) for (@ret);
  return @ret;
}

=item tag_list_length()

returns the length of this TagList

=cut
sub tag_list_length
{ return scalar(shift->key_list()); }

=item has_tag(RegExp)

return true if the given TagName exists

=cut
sub has_tag
{ return has_key(@_); }

=item has_tags()

return true if Tags exists

=cut
sub has_tags
{ return shift->key_list_length()>0; }

=item tag_hash(name1=>key1,name2=>key2,name3=>key3)

return a hash with the found keys and names as key

Example:

 %tags=$tags->tag_hash(Name=>'name', BusLinie=>'line');
 # returns tomthing like that:
 %tags=(
   Name=>... OSM::Tree::Tag - Object
   BusLinie=>... OSM::Tree::Tag - Object
 );

=cut
sub tags_hash
{
  my $self=shift;
  return () if(@_%2);
  my %in=@_;
  my %ret=();
  while (my ($key,$val)=each(%in))
  {
    if($self->has_key($val))
    { $ret{$key}=$self->tag($val); }
  }
  return wantarray?%ret:\%ret;
}

sub to_xml
{
  my $self=shift;
  my $in=shift || '';
  my $ret='';
  $ret.=$_->to_xml($in) for(sort{$a->name() cmp $b->name()}$self->tag_list());
  return $ret;
}
########################################################################

=back

=head2 OSM::Tree::Tag

the Tag-Object

overloads "" and cmp

=head1 USING

L<overload>

=over 4

=cut

package OSM::Tree::Tag;
use overload  '""'  => sub{ return shift->value(); },
              'cmp' => \&mycmp;

sub mycmp
{
  my $val1=shift->value();
  my $val2=shift;
  my $flag=shift;
  my $result;
  if ((defined $flag) && $flag)
  { $result = $val2 cmp $val1; }
  else
  { $result = $val1 cmp $val2; }
  return $result;
};

sub _new
{
  my $class=shift;
  my $self={};
  $self->{ELEMENT}=shift;
  $self->{NAME}=shift;
  $self->{VALUE}=__parse(shift);
  bless($self,$class);
  return $self;
}

sub __parse
{
  my $value=shift;
  return $value if(ref($value) eq 'ARRAY');
  return [] unless defined($value);
  if($value=~/;/)
  {
    $value=~s/^\s*;?\s*(.+?)\s*;?\s*$/$1/gs;
    return [split(/\s*;\s*/,$value)];
  }
  return [$value];
}

#=======================================================================
=item element()

returns the parent object, of this tag-object;
can be: node, way or relation

=cut
sub element
{ return shift->{ELEMENT}; }

=item name()

returns the name of this Tag

=cut
sub name
{ return shift->{NAME}; }

#=======================================================================

=item value([POS])

return the joined values if no Parameter ist set
or
returns the parameter at the Position pased trou the Parameter
or
returns undef if the asked Value do not exists

Example:
 $first_value=$tag->value(0);
 $last_value =$tag->value(-1);
 $third_value=$tag->value(2);
 $all_values_joined=$tag->value();
=cut
sub value
{
  my $self=shift;
  my $pos=shift;
  return $self->_value_list_joined() unless defined($pos);
  return $self->{VALUE}->[$pos];
}

sub _value_list_joined
{
  my $self=shift;
  my $sep=shift || ';';
  return join($sep,$self->value_list());
}

=item value_set(VALUE)

returns 1 on success

=cut
sub value_set
{
  my $self=shift;
  my $value=shift;
  return 0 unless(defined($value));
  $self->{VALUE}=__parse($value);
  $self->element->_set_modify();
  return 1;
}

=item value_list()

returns an List of the Values of this Tag

=cut
sub value_list
{ return @{shift->{VALUE}}; }

=item value_list_length()

returns the length valuelist of this Tag

=cut
sub value_list_length
{ return scalar(shift->value_list()); }

#=======================================================================

=item value_inhertit_one(@LIST)

returns 1 if at least one Element of the List is an Value of this Tag
otherwise 0 will be returned

=cut
sub value_inherit_one
{
  my $self=shift;
  my @list=@_;
  for my $elm ($self->value_list())
  { return 1 if(grep{$elm eq $_}@list); }
  return 0;
}

=item value_inhertit_all(@LIST)

returns 1 if at least one Element of the List is an Value of this Tag
otherwise 0 will be returned

=cut
sub value_inherit_all
{
  my $self=shift;
  my @list=@_;
  my $ret=0;
  for my $elm (@list)
  { $ret++ if(grep{$elm eq $_}$self->value_list()); }
  return @list==$ret;
}

=item value_match(RegExp)

returns a list of Values who matchs the ReguarExpression

=cut
sub value_match
{
  my $self=shift;
  my $regexp=shift;
  my @ret;
  for my $elm ($self->value_list())
  { push(@ret,$elm) if($elm=~/$regexp/); }
  return @ret;
}

sub to_xml
{
  my $self=shift;
  my $in=shift || '';
  my $k=$self->name();
  my $v=$self->value();
  return qq!$in<tag k="$k" v="$v"/>\n!;
}
########################################################################
package OSM::Tree::TagDummy;
use overload  '""'  => sub{ return shift->value(); },
              'cmp' => \&mycmp;

sub mycmp
{
  my $val1=shift->value();
  my $val2=shift;
  my $flag=shift;
  my $result;
  if ((defined $flag) && $flag)
  { $result = $val2 cmp $val1; }
  else
  { $result = $val1 cmp $val2; }
  return $result;
};

sub _new
{
  my $class=shift;
  my $self={};
  $self->{ELEMENT}=shift;
  $self->{NAME}=shift;
  bless($self,$class);
  return $self;
}

sub element { return shift->{ELEMENT}; }
sub name { return shift->{NAME}; }

sub value
{
  return '' if(@_==1);
  return undef;
}
sub _value_list_joined { return ''; }
sub value_list { return (); }
sub value_list_length { return 0; }
sub value_inherit_one { return 0; }
sub value_inherit_all { return 0; }
sub value_match { return (); }
sub value_set { return 0; }

sub to_xml
{
  my $self=shift;
  my $in=shift || '';
  my $k=$self->name();
  return qq!$in<tag k="$k" v=""/>\n!;
}


1;


=back

=head1 See also

L<OSM::Tree::BASE>

L<OSM::Tree::Node>

L<OSM::Tree::Way>

L<OSM::Tree::Relation>

L<OSM::Tree::TagList>

L<OSM::Tree::Polygon>

L<OSM::Tree>

=cut
